import os
from typing import List, Dict, Any

import yaml
from jinja2 import Environment, FileSystemLoader, select_autoescape, StrictUndefined

TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), "templates")


def render_config(
    helm_config_path: str,
    output_path: str,
    slug: str,
    environment: str,
    params: Dict[str, Any],
) -> None:
    # get helm data
    configmaps = get_helm_data(helm_config_path, file_type="configmap")
    deployments = get_helm_data(helm_config_path, file_type="deployment")
    env = get_helm_data(helm_config_path, file_type="deployment", index=True)

    # extract envs
    envs = extract_env_vars(env)

    docker_image_versions = dict(
        yaml_open(d)["spec"]["template"]["spec"]["containers"][0]["image"].split(":")
        for d in deployments
    )

    # merge envs and params
    merged_params = {
        **params,
        **envs,
        "slug": slug,
        "environment": environment,
        "docker_image_versions": docker_image_versions,
    }

    # prepare the configs
    mapping = prepare_configs(configmaps)

    # create folder with name of slug at path
    destination = prepare_directory(
        output_path,
        slug,
        environment=environment,
    )

    # create necessary configs and store at appropriate path
    create_configurations(destination, mapping)

    # fill remaining templates and store at appropriate path
    render_templates(destination, merged_params)


def get_helm_data(helm_config_path: str, file_type: str, index: bool = False) -> Any:
    if not os.path.isdir(helm_config_path):
        raise ValueError(f"{helm_config_path} is not a directory")

    to_remove = ["redis", "renderer", "database"]
    rendered_templates = os.listdir(helm_config_path)

    for template in to_remove:
        rendered_templates.remove(template)

    files = []
    for template_dir in rendered_templates:
        template_path = os.path.join(helm_config_path, template_dir)
        template_path = os.path.join(template_path, "templates")
        for f in os.listdir(template_path):
            if file_type in f and os.path.isfile(
                template_file := os.path.join(template_path, f)
            ):
                files.append(template_file)

    if index:
        for i, f in enumerate(files):
            if "registrar" in f:
                return files[i]

    return files


def prepare_configs(configmaps: List[str]) -> Dict[str, str]:
    file_config_mapping = {}
    for configmap in configmaps:
        config = yaml_open(configmap)
        config = config["data"]
        if config:
            file_name = list(config.keys())[0]
            config = config[file_name]
        else:
            continue

        file_config_mapping[file_name] = config

    return file_config_mapping


def prepare_directory(output_path: str, slug: str, environment: str) -> str:
    if os.path.isdir(output_path):
        destination = os.path.join(output_path, f"{slug}-{environment}")
        try:
            os.mkdir(destination)
        except FileExistsError:
            pass
    else:
        raise ValueError(f"{output_path} is not a directory")

    return destination


def create_configurations(destination: str, mapping: Dict[str, str]) -> None:
    destination = os.path.join(destination, "config")
    try:
        os.mkdir(destination)
    except FileExistsError:
        pass

    for filename, config in mapping.items():
        config_path = os.path.join(destination, filename)
        with open(config_path, "w", encoding="utf-8") as cfg:
            cfg.write(config)


def render_templates(destination: str, params: Dict[str, Any]) -> None:
    jinja_env = Environment(
        loader=FileSystemLoader(TEMPLATE_PATH),
        autoescape=select_autoescape(),
        undefined=StrictUndefined,
    )

    templates = os.listdir(TEMPLATE_PATH)
    for template in templates:
        tpl = jinja_env.get_template(template)
        rendered_template = tpl.render(**params)
        template_destination = (
            os.path.join(destination, template)
            if "docker" in template
            else os.path.join(destination, "config", template)
        )
        with open(template_destination, "w", encoding="utf-8") as tpl_file:
            tpl_file.write(rendered_template)


def extract_env_vars(env: str) -> Dict[str, str]:
    envs = {}
    config = yaml_open(env)
    env_vars = config["spec"]["template"]["spec"]["containers"][0]["env"]
    for var in env_vars:
        envs[var["name"]] = var["value"]

    return envs


def yaml_open(path: str) -> Any:
    with open(path, "r", encoding="utf-8") as f:
        return yaml.safe_load(f)
